@@ -1,7 +1,9 @@
 import re
 import ssl
+from contextlib import contextmanager
 from time import time
 import types
+from urllib.parse import urlparse, parse_qs, unquote
 
 from . import errors, defines
 from .block import ColumnOrientedBlock, RowOrientedBlock
@@ -11,9 +13,8 @@ from .protocol import ServerPacketTypes
 from .result import (
     IterQueryResult, ProgressQueryResult, QueryResult, QueryInfo
 )
-from .util.compat import urlparse, parse_qs, asbool, unquote
 from .util.escape import escape_params
-from .util.helpers import column_chunks, chunks
+from .util.helpers import column_chunks, chunks, asbool
 
 
 class Client(object):
@@ -21,33 +22,47 @@ class Client(object):
     Client for communication with the ClickHouse server.
     Single connection is established per each connected instance of the client.
 
-    :param settings: Dictionary of settings that passed to every query.
-                     Defaults to ``None`` (no additional settings). See all
-                     available settings in `ClickHouse docs
-                     <https://clickhouse.yandex/docs/en/single/#settings>`_.
+    :param settings: Dictionary of settings that passed to every query (except
+                     for the client settings, see below). Defaults to ``None``
+                     (no additional settings). See all available settings in
+                     `ClickHouse docs
+                     <https://clickhouse.tech/docs/en/operations/settings/settings/>`_.
+    :param \\**kwargs: All other args are passed to the
+                       :py:class:`~clickhouse_driver.connection.Connection`
+                       constructor.
 
-    Driver's settings:
+    The following keys when passed in ``settings`` are used for configuring the
+    client itself:
 
-        * insert_block_size -- chunk size to split rows for ``INSERT``.
+        * ``insert_block_size`` -- chunk size to split rows for ``INSERT``.
           Defaults to ``1048576``.
-
-        * strings_as_bytes -- turns off string column encoding/decoding.
-
-        * strings_encoding -- specifies string encoding. UTF-8 by default.
-
-        * use_numpy -- Use numpy for columns reading. New in version *0.2.0*.
-
+        * ``strings_as_bytes`` -- turns off string column encoding/decoding.
+        * ``strings_encoding`` -- specifies string encoding. UTF-8 by default.
+        * ``use_numpy`` -- Use numpy for columns reading. New in version
+                           *0.2.0*.
+        * ``opentelemetry_traceparent`` -- OpenTelemetry traceparent header as
+                           described by W3C Trace Context recommendation.
+                           New in version *0.2.2*.
+        * ``opentelemetry_tracestate`` -- OpenTelemetry tracestate header as
+                           described by W3C Trace Context recommendation.
+                           New in version *0.2.2*.
+        * ``quota_key`` -- A string to differentiate quotas when the user have
+                           keyed quotas configured on server.
+                           New in version *0.2.3*.
     """
 
     available_client_settings = (
         'insert_block_size',  # TODO: rename to max_insert_block_size
         'strings_as_bytes',
         'strings_encoding',
-        'use_numpy'
+        'use_numpy',
+        'opentelemetry_traceparent',
+        'opentelemetry_tracestate',
+        'quota_key'
     )
 
     def __init__(self, *args, **kwargs):
-        self.settings = kwargs.pop('settings', {}).copy()
+        self.settings = (kwargs.pop('settings', None) or {}).copy()
 
         self.client_settings = {
             'insert_block_size': int(self.settings.pop(
@@ -61,6 +76,15 @@ class Client(object):
             ),
             'use_numpy': self.settings.pop(
                 'use_numpy', False
+            ),
+            'opentelemetry_traceparent': self.settings.pop(
+                'opentelemetry_traceparent', None
+            ),
+            'opentelemetry_tracestate': self.settings.pop(
+                'opentelemetry_tracestate', ''
+            ),
+            'quota_key': self.settings.pop(
+                'quota_key', ''
             )
         }
 
@@ -86,6 +110,12 @@ class Client(object):
         self.reset_last_query()
         super(Client, self).__init__()
 
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.disconnect()
+
     def disconnect(self):
         """
         Disconnects from the server.
@@ -184,6 +214,27 @@ class Client(object):
         query_settings.update(settings)
         self.connection.context.settings = query_settings
 
+    def track_current_database(self, query):
+        query = query.strip('; ')
+        if query.lower().startswith('use '):
+            self.connection.database = query[4:].strip()
+
+    @contextmanager
+    def disconnect_on_error(self, query, settings):
+        self.make_query_settings(settings)
+
+        try:
+            self.connection.force_connect()
+            self.last_query = QueryInfo()
+
+            yield
+
+            self.track_current_database(query)
+
+        except (Exception, KeyboardInterrupt):
+            self.disconnect()
+            raise
+
     def execute(self, query, params=None, with_column_types=False,
                 external_tables=None, query_id=None, settings=None,
                 types_check=False, columnar=False):
@@ -229,11 +280,8 @@ class Client(object):
         """
 
         start_time = time()
-        self.make_query_settings(settings)
-        self.connection.force_connect()
-        self.last_query = QueryInfo()
 
-        try:
+        with self.disconnect_on_error(query, settings):
             # INSERT queries can use list/tuple/generator of list/tuples/dicts.
             # For SELECT parameters can be passed in only in dict right now.
             is_insert = isinstance(params, (list, tuple, types.GeneratorType))
@@ -254,10 +302,6 @@ class Client(object):
             self.last_query.store_elapsed(time() - start_time)
             return rv
 
-        except (Exception, KeyboardInterrupt):
-            self.disconnect()
-            raise
-
     def execute_with_progress(
             self, query, params=None, with_column_types=False,
             external_tables=None, query_id=None, settings=None,
@@ -288,21 +332,13 @@ class Client(object):
         :return: :ref:`progress-query-result` proxy.
         """
 
-        self.make_query_settings(settings)
-        self.connection.force_connect()
-        self.last_query = QueryInfo()
-
-        try:
+        with self.disconnect_on_error(query, settings):
             return self.process_ordinary_query_with_progress(
                 query, params=params, with_column_types=with_column_types,
                 external_tables=external_tables, query_id=query_id,
                 types_check=types_check, columnar=columnar
             )
 
-        except (Exception, KeyboardInterrupt):
-            self.disconnect()
-            raise
-
     def execute_iter(
             self, query, params=None, with_column_types=False,
             external_tables=None, query_id=None, settings=None,
@@ -331,21 +367,13 @@ class Client(object):
         :return: :ref:`iter-query-result` proxy.
         """
 
-        self.make_query_settings(settings)
-        self.connection.force_connect()
-        self.last_query = QueryInfo()
-
-        try:
+        with self.disconnect_on_error(query, settings):
             return self.iter_process_ordinary_query(
                 query, params=params, with_column_types=with_column_types,
                 external_tables=external_tables,
                 query_id=query_id, types_check=types_check
             )
 
-        except (Exception, KeyboardInterrupt):
-            self.disconnect()
-            raise
-
     def query_dataframe(
             self, query, params=None, external_tables=None, query_id=None,
             settings=None):
@@ -382,8 +410,8 @@ class Client(object):
         )
 
     def insert_dataframe(
-            self, query, dataframe, transpose=True, external_tables=None,
-            query_id=None, settings=None):
+            self, query, dataframe, external_tables=None, query_id=None,
+            settings=None):
         """
         *New in version 0.2.0.*
 
@@ -391,11 +419,6 @@ class Client(object):
 
         :param query: query that will be send to server.
         :param dataframe: pandas DataFrame.
-        :param transpose: whether or not transpose DataFrame before sending.
-                          This is necessary action as DataFrame can be sent in
-                          columnar form. If DataFrame is already in columnar
-                          form set this parameter to ``False``.
-                          Defaults to ``True``.
         :param external_tables: external tables to send.
                                 Defaults to ``None`` (no external tables).
         :param query_id: the query identifier. If no query id specified
@@ -410,13 +433,22 @@ class Client(object):
         except ImportError:
             raise RuntimeError('Extras for NumPy must be installed')
 
-        frame = dataframe.transpose() if transpose else dataframe
-        columns = list(frame.values)
+        start_time = time()
 
-        return self.execute(
-            query, columns, columnar=True, external_tables=external_tables,
-            query_id=query_id, settings=settings
-        )
+        with self.disconnect_on_error(query, settings):
+            self.connection.send_query(query, query_id=query_id)
+            self.connection.send_external_tables(external_tables)
+
+            sample_block = self.receive_sample_block()
+            rv = None
+            if sample_block:
+                columns = [x[0] for x in sample_block.columns_with_types]
+                data = [dataframe[column].values for column in columns]
+                rv = self.send_data(sample_block, data, columnar=True)
+                self.receive_end_of_query()
+
+            self.last_query.store_elapsed(time() - start_time)
+            return rv
 
     def process_ordinary_query_with_progress(
             self, query, params=None, with_column_types=False,
@@ -424,7 +456,9 @@ class Client(object):
             types_check=False, columnar=False):
 
         if params is not None:
-            query = self.substitute_params(query, params)
+            query = self.substitute_params(
+                query, params, self.connection.context
+            )
 
         self.connection.send_query(query, query_id=query_id)
         self.connection.send_external_tables(external_tables,
@@ -438,7 +472,9 @@ class Client(object):
             types_check=False, columnar=False):
 
         if params is not None:
-            query = self.substitute_params(query, params)
+            query = self.substitute_params(
+                query, params, self.connection.context
+            )
 
         self.connection.send_query(query, query_id=query_id)
         self.connection.send_external_tables(external_tables,
@@ -452,7 +488,9 @@ class Client(object):
             types_check=False):
 
         if params is not None:
-            query = self.substitute_params(query, params)
+            query = self.substitute_params(
+                query, params, self.connection.context
+            )
 
         self.connection.send_query(query, query_id=query_id)
         self.connection.send_external_tables(external_tables,
@@ -556,11 +594,11 @@ class Client(object):
         # Client must still read until END_OF_STREAM packet.
         return self.receive_result(with_column_types=with_column_types)
 
-    def substitute_params(self, query, params):
+    def substitute_params(self, query, params, context):
         if not isinstance(params, dict):
             raise ValueError('Parameters are expected in dict form')
 
-        escaped = escape_params(params)
+        escaped = escape_params(params, context)
         return query % escaped
 
     @classmethod
@@ -646,7 +684,7 @@ class Client(object):
                 kwargs[name] = asbool(value)
             elif name == 'ssl_version':
                 kwargs[name] = getattr(ssl, value)
-            elif name in ['ca_certs', 'ciphers']:
+            elif name in ['ca_certs', 'ciphers', 'keyfile', 'certfile']:
                 kwargs[name] = value
             elif name == 'alt_hosts':
                 kwargs['alt_hosts'] = value
