@@ -4,6 +4,7 @@ import ssl
 from collections import deque
 from contextlib import contextmanager
 from time import time
+from urllib.parse import urlparse
 
 from . import defines
 from . import errors
@@ -21,7 +22,8 @@ from .queryprocessingstage import QueryProcessingStage
 from .reader import read_binary_str
 from .readhelpers import read_exception
 from .settings.writer import write_settings
-from .util.compat import urlparse
+from .streams.native import BlockInputStream, BlockOutputStream
+from .util.compat import threading
 from .varint import write_varint, read_varint
 from .writer import write_binary_str
 
@@ -109,6 +111,8 @@ class Connection(object):
     :param ssl_version: see :func:`ssl.wrap_socket` docs.
     :param ca_certs: see :func:`ssl.wrap_socket` docs.
     :param ciphers: see :func:`ssl.wrap_socket` docs.
+    :param keyfile: see :func:`ssl.wrap_socket` docs.
+    :param certfile: see :func:`ssl.wrap_socket` docs.
     :param alt_hosts: list of alternative hosts for connection.
                       Example: alt_hosts=host1:port1,host2:port2.
     :param settings_is_important: ``False`` means unknown settings will be
@@ -119,7 +123,8 @@ class Connection(object):
 
     def __init__(
             self, host, port=None,
-            database='default', user='default', password='',
+            database=defines.DEFAULT_DATABASE,
+            user=defines.DEFAULT_USER, password=defines.DEFAULT_PASSWORD,
             client_name=defines.CLIENT_NAME,
             connect_timeout=defines.DBMS_DEFAULT_CONNECT_TIMEOUT_SEC,
             send_receive_timeout=defines.DBMS_DEFAULT_TIMEOUT_SEC,
@@ -129,6 +134,7 @@ class Connection(object):
             secure=False,
             # Secure socket parameters.
             verify=True, ssl_version=None, ca_certs=None, ciphers=None,
+            keyfile=None, certfile=None,
             alt_hosts=None,
             settings_is_important=False,
     ):
@@ -163,6 +169,10 @@ class Connection(object):
             ssl_options['ca_certs'] = ca_certs
         if ciphers is not None:
             ssl_options['ciphers'] = ciphers
+        if keyfile is not None:
+            ssl_options['keyfile'] = keyfile
+        if certfile is not None:
+            ssl_options['certfile'] = certfile
 
         self.ssl_options = ssl_options
 
@@ -185,12 +195,17 @@ class Connection(object):
 
         self.connected = False
 
+        self.client_trace_context = None
         self.server_info = None
         self.context = Context()
 
         # Block writer/reader
         self.block_in = None
         self.block_out = None
+        self.block_in_raw = None  # log blocks are always not compressed
+
+        self._lock = threading.Lock()
+        self.is_query_executing = False
 
         super(Connection, self).__init__()
 
@@ -198,6 +213,8 @@ class Connection(object):
         return '{}:{}'.format(self.host, self.port)
 
     def force_connect(self):
+        self.check_query_execution()
+
         if not self.connected:
             self.connect()
 
@@ -262,6 +279,10 @@ class Connection(object):
         if 'cert_reqs' in ssl_options:
             context.options = ssl_options['cert_reqs']
 
+        if 'certfile' in ssl_options:
+            keyfile = ssl_options.get('keyfile')
+            context.load_cert_chain(ssl_options['certfile'], keyfile=keyfile)
+
         return context
 
     def _init_connection(self, host, port):
@@ -280,6 +301,7 @@ class Connection(object):
         self.receive_hello()
 
         self.block_in = self.get_block_in_stream()
+        self.block_in_raw = BlockInputStream(self.fin, self.context)
         self.block_out = self.get_block_out_stream()
 
     def _format_connection_error(self, e, host, port):
@@ -332,11 +354,15 @@ class Connection(object):
 
         self.connected = False
 
+        self.client_trace_context = None
         self.server_info = None
 
         self.block_in = None
+        self.block_in_raw = None
         self.block_out = None
 
+        self.is_query_executing = False
+
     def disconnect(self):
         """
         Closes connection between server and client.
@@ -415,9 +441,9 @@ class Connection(object):
             raise self.receive_exception()
 
         else:
-            self.disconnect()
             message = self.unexpected_packet_message('Hello or Exception',
                                                      packet_type)
+            self.disconnect()
             raise errors.UnexpectedPacketFromServerError(message)
 
     def ping(self):
@@ -461,10 +487,10 @@ class Connection(object):
         elif packet_type == ServerPacketTypes.EXCEPTION:
             packet.exception = self.receive_exception()
 
-        elif packet.type == ServerPacketTypes.PROGRESS:
+        elif packet_type == ServerPacketTypes.PROGRESS:
             packet.progress = self.receive_progress()
 
-        elif packet.type == ServerPacketTypes.PROFILE_INFO:
+        elif packet_type == ServerPacketTypes.PROFILE_INFO:
             packet.profile_info = self.receive_profile_info()
 
         elif packet_type == ServerPacketTypes.TOTALS:
@@ -474,10 +500,11 @@ class Connection(object):
             packet.block = self.receive_data()
 
         elif packet_type == ServerPacketTypes.LOG:
-            block = self.receive_data()
+            block = self.receive_data(raw=True)
             log_block(block)
 
         elif packet_type == ServerPacketTypes.END_OF_STREAM:
+            self.is_query_executing = False
             pass
 
         elif packet_type == ServerPacketTypes.TABLE_COLUMNS:
@@ -486,12 +513,11 @@ class Connection(object):
             )
 
         else:
-            self.disconnect()
-            raise errors.UnknownPacketFromServerError(
-                'Unknown packet {} from server {}'.format(
-                    packet_type, self.get_description()
-                )
+            message = 'Unknown packet {} from server {}'.format(
+                packet_type, self.get_description()
             )
+            self.disconnect()
+            raise errors.UnknownPacketFromServerError(message)
 
         return packet
 
@@ -501,8 +527,6 @@ class Connection(object):
 
             return CompressedBlockInputStream(self.fin, self.context)
         else:
-            from .streams.native import BlockInputStream
-
             return BlockInputStream(self.fin, self.context)
 
     def get_block_out_stream(self):
@@ -514,17 +538,15 @@ class Connection(object):
                 self.fout, self.context
             )
         else:
-            from .streams.native import BlockOutputStream
-
             return BlockOutputStream(self.fout, self.context)
 
-    def receive_data(self):
+    def receive_data(self, raw=False):
         revision = self.server_info.revision
 
         if revision >= defines.DBMS_MIN_REVISION_WITH_TEMPORARY_TABLES:
             read_binary_str(self.fin)
 
-        return self.block_in.read()
+        return (self.block_in_raw if raw else self.block_in).read()
 
     def receive_exception(self):
         return read_exception(self.fin)
@@ -564,7 +586,7 @@ class Connection(object):
 
         revision = self.server_info.revision
         if revision >= defines.DBMS_MIN_REVISION_WITH_CLIENT_INFO:
-            client_info = ClientInfo(self.client_name)
+            client_info = ClientInfo(self.client_name, self.context)
             client_info.query_kind = ClientInfo.QueryKind.INITIAL_QUERY
 
             client_info.write(revision, self.fout)
@@ -595,8 +617,24 @@ class Connection(object):
 
     def send_external_tables(self, tables, types_check=False):
         for table in tables or []:
-            block = RowOrientedBlock(table['structure'], table['data'],
-                                     types_check=types_check)
+            if not table['structure']:
+                raise ValueError(
+                    'Empty table "{}" structure'.format(table['name'])
+                )
+
+            data = table['data']
+            block_cls = RowOrientedBlock
+
+            if self.context.client_settings['use_numpy']:
+                from .numpy.block import NumpyColumnOrientedBlock
+
+                columns = [x[0] for x in table['structure']]
+                data = [data[column].values for column in columns]
+
+                block_cls = NumpyColumnOrientedBlock
+
+            block = block_cls(table['structure'], data,
+                              types_check=types_check)
             self.send_data(block, table_name=table['name'])
 
         # Empty block, end of data transfer.
@@ -618,3 +656,12 @@ class Connection(object):
             'Unexpected packet from server {} (expected {}, got {})'
             .format(self.get_description(), expected, packet_type)
         )
+
+    def check_query_execution(self):
+        self._lock.acquire(blocking=False)
+
+        if self.is_query_executing:
+            raise errors.PartiallyConsumedQueryError()
+
+        self.is_query_executing = True
+        self._lock.release()
