@@ -1,4 +1,8 @@
 
+DEFAULT_DATABASE = 'default'
+DEFAULT_USER = 'default'
+DEFAULT_PASSWORD = ''
+
 DEFAULT_PORT = 9000
 DEFAULT_SECURE_PORT = 9440
 
@@ -16,6 +20,7 @@ DBMS_MIN_REVISION_WITH_COLUMN_DEFAULTS_METADATA = 54410
 DBMS_MIN_REVISION_WITH_CLIENT_WRITE_INFO = 54420
 DBMS_MIN_REVISION_WITH_SETTINGS_SERIALIZED_AS_STRINGS = 54429
 DBMS_MIN_REVISION_WITH_INTERSERVER_SECRET = 54441
+DBMS_MIN_REVISION_WITH_OPENTELEMETRY = 54442
 
 # Timeouts
 DBMS_DEFAULT_CONNECT_TIMEOUT_SEC = 10
@@ -31,7 +36,7 @@ CLIENT_NAME = 'python-driver'
 CLIENT_VERSION_MAJOR = 20
 CLIENT_VERSION_MINOR = 10
 CLIENT_VERSION_PATCH = 2
-CLIENT_REVISION = 54441
+CLIENT_REVISION = 54442
 
 BUFFER_SIZE = 1048576
 
